import os
import sys
import io
from contextlib import contextmanager

from cython.operator cimport dereference
from libc.stdio cimport stdout
from libc.stdlib cimport malloc, free
from libc.string cimport strcmp
from numpy cimport ndarray
ctypedef double complex double_complex

from pyquda_comm.pointer cimport Pointer, Pointers, _NDArray
cimport quda


@contextmanager
def redirect_stdout(value: bytearray):
    stdout_fd = sys.stdout.fileno()
    stdout_dup_fd = os.dup(stdout_fd)
    pipe_out, pipe_in = os.pipe()
    os.dup2(pipe_in, stdout_fd)

    yield

    sys.stdout.write(b"\x00".decode(sys.stdout.encoding))
    sys.stdout.flush()
    os.close(pipe_in)
    with io.FileIO(pipe_out, closefd=True) as fio:
        buffer = fio.read(4096)
        while b"\00" not in buffer:
            value.extend(buffer)
            buffer = fio.read(4096)
        value.extend(buffer)
    os.dup2(stdout_dup_fd, stdout_fd)
    os.close(stdout_dup_fd)


cdef class QudaGaugeParam:
    cdef quda.QudaGaugeParam param

    def __init__(self):
        self.param = quda.newQudaGaugeParam()

    def __repr__(self):
        value = bytearray()
        with redirect_stdout(value):
            quda.printQudaGaugeParam(&self.param)
        return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaGaugeParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaGaugeParam

cdef class QudaInvertParam:
    cdef quda.QudaInvertParam param

    def __init__(self):
        self.param = quda.newQudaInvertParam()

    def __repr__(self):
        value = bytearray()
        with redirect_stdout(value):
            quda.printQudaInvertParam(&self.param)
        return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaInvertParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaInvertParam

cdef class QudaMultigridParam:
    cdef quda.QudaMultigridParam param

    def __init__(self):
        self.param = quda.newQudaMultigridParam()

    def __repr__(self):
        value = bytearray()
        with redirect_stdout(value):
            quda.printQudaMultigridParam(&self.param)
        return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaMultigridParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaMultigridParam

cdef class QudaEigParam:
    cdef quda.QudaEigParam param

    def __init__(self):
        self.param = quda.newQudaEigParam()

    def __repr__(self):
        value = bytearray()
        with redirect_stdout(value):
            quda.printQudaEigParam(&self.param)
        return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaEigParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaEigParam

cdef class QudaGaugeObservableParam:
    cdef quda.QudaGaugeObservableParam param

    def __init__(self):
        self.param = quda.newQudaGaugeObservableParam()

    def __repr__(self):
        value = bytearray()
        with redirect_stdout(value):
            quda.printQudaGaugeObservableParam(&self.param)
        return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaGaugeObservableParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaGaugeObservableParam

cdef class QudaGaugeSmearParam:
    cdef quda.QudaGaugeSmearParam param

    def __init__(self):
        self.param = quda.newQudaGaugeSmearParam()

    # def __repr__(self):
    #     value = bytearray()
    #     with redirect_stdout(value):
    #         quda.printQudaGaugeSmearParam(&self.param)
    #     return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaGaugeSmearParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaGaugeSmearParam

cdef class QudaBLASParam:
    cdef quda.QudaBLASParam param

    def __init__(self):
        self.param = quda.newQudaBLASParam()

    def __repr__(self):
        value = bytearray()
        with redirect_stdout(value):
            quda.printQudaBLASParam(&self.param)
        return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaBLASParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaBLASParam

def setVerbosityQuda(quda.QudaVerbosity verbosity, const char prefix[]):
    quda.setVerbosityQuda(verbosity, prefix, stdout)

ctypedef struct CommsMapData:
    int ndim
    int dims[6]
    int size
    int *ranks

cdef CommsMapData map_data

cdef int commsMap(const int *coords, void *fdata) noexcept:
    cdef CommsMapData *md = <CommsMapData *>fdata

    cdef int rank = 0
    for i in range(md.ndim):
        rank = rank * md.dims[i] + coords[i]

    for i in range(md.size):
        if rank == md.ranks[i]:
            return i
    else:
        return rank

def initCommsGridQuda(int nDim, list dims, list ranks):
    map_data.ndim = nDim
    map_data.size = 1
    for i in range(map_data.ndim):
        map_data.dims[i] = dims[i]
        map_data.size *= dims[i]
    map_data.ranks = <int *>malloc(map_data.size * sizeof(int))
    for i in range(map_data.size):
        map_data.ranks[i] = ranks[i]
    quda.initCommsGridQuda(map_data.ndim, map_data.dims, commsMap, <void *>(&map_data))

def initQudaDevice(int device):
    quda.initQudaDevice(device)

def initQudaMemory():
    quda.initQudaMemory()

def initQuda(int device):
    quda.initQuda(device)

def endQuda():
    free(map_data.ranks)
    quda.endQuda()

def updateR():
    quda.updateR()

def loadGaugeQuda(h_gauge, QudaGaugeParam param):
    _h_gauge = _NDArray(h_gauge, 2)
    quda.loadGaugeQuda(_h_gauge.ptr, &param.param)

def freeGaugeQuda():
    quda.freeGaugeQuda()

def freeUniqueGaugeQuda(quda.QudaLinkType link_type):
    quda.freeUniqueGaugeQuda(link_type)

def freeGaugeSmearedQuda():
    quda.freeGaugeSmearedQuda()

def saveGaugeQuda(h_gauge, QudaGaugeParam param):
    _h_gauge = _NDArray(h_gauge, 2)
    quda.saveGaugeQuda(_h_gauge.ptr, &param.param)

def loadCloverQuda(h_clover, h_clovinv, QudaInvertParam inv_param):
    _h_clover = _NDArray(h_clover, 1)
    _h_clovinv = _NDArray(h_clovinv, 1)
    quda.loadCloverQuda(_h_clover.ptr, _h_clovinv.ptr, &inv_param.param)

def freeCloverQuda():
    quda.freeCloverQuda()

# QUDA only declares lanczosQuda
# def lanczosQuda(int k0, int m, Pointer hp_Apsi, Pointer hp_r, Pointer hp_V, Pointer hp_alpha, Pointer hp_beta, QudaEigParam eig_param):

def eigensolveQuda(h_evecs, ndarray[double_complex, ndim=1] h_evals, QudaEigParam param):
    _h_evecs = _NDArray(h_evecs, 2)
    _h_evals = _NDArray(h_evals)
    quda.eigensolveQuda(_h_evecs.ptrs, <double_complex *>_h_evals.ptr, &param.param)

def invertQuda(h_x, h_b, QudaInvertParam param):
    _h_x = _NDArray(h_x, 1)
    _h_b = _NDArray(h_b, 1)
    quda.invertQuda(_h_x.ptr, _h_b.ptr, &param.param)

def invertMultiSrcQuda(_hp_x, _hp_b, QudaInvertParam param):
    __hp_x = _NDArray(_hp_x, 2)
    __hp_b = _NDArray(_hp_b, 2)
    quda.invertMultiSrcQuda(__hp_x.ptrs, __hp_b.ptrs, &param.param)

def invertMultiShiftQuda(_hp_x, _hp_b, QudaInvertParam param):
    __hp_x = _NDArray(_hp_x, 2)
    __hp_b = _NDArray(_hp_b, 1)
    quda.invertMultiShiftQuda(__hp_x.ptrs, __hp_b.ptr, &param.param)

def newMultigridQuda(QudaMultigridParam param):
    mg_instance = Pointer("void")
    mg_instance.set_ptr(quda.newMultigridQuda(&param.param))
    return mg_instance

def destroyMultigridQuda(Pointer mg_instance):
    quda.destroyMultigridQuda(mg_instance.ptr)

def updateMultigridQuda(Pointer mg_instance, QudaMultigridParam param):
    quda.updateMultigridQuda(mg_instance.ptr, &param.param)

def dumpMultigridQuda(Pointer mg_instance, QudaMultigridParam param):
    quda.dumpMultigridQuda(mg_instance.ptr, &param.param)

def dslashQuda(h_out, h_in, QudaInvertParam inv_param, quda.QudaParity parity):
    _h_out = _NDArray(h_out, 1)
    _h_in = _NDArray(h_in, 1)
    quda.dslashQuda(_h_out.ptr, _h_in.ptr, &inv_param.param, parity)

def dslashMultiSrcQuda(_hp_x, _hp_b, QudaInvertParam param, quda.QudaParity parity):
    __hp_x = _NDArray(_hp_x, 2)
    __hp_b = _NDArray(_hp_b, 2)
    quda.dslashMultiSrcQuda(__hp_x.ptrs, __hp_b.ptrs, &param.param, parity)

def cloverQuda(h_out, h_in, QudaInvertParam inv_param, quda.QudaParity parity, int inverse):
    _h_out = _NDArray(h_out, 1)
    _h_in = _NDArray(h_in, 1)
    quda.cloverQuda(_h_out.ptr, _h_in.ptr, &inv_param.param, parity, inverse)

def MatQuda(h_out, h_in, QudaInvertParam inv_param):
    _h_out = _NDArray(h_out, 1)
    _h_in = _NDArray(h_in, 1)
    quda.MatQuda(_h_out.ptr, _h_in.ptr, &inv_param.param)

def MatDagMatQuda(h_out, h_in, QudaInvertParam inv_param):
    _h_out = _NDArray(h_out, 1)
    _h_in = _NDArray(h_in, 1)
    quda.MatDagMatQuda(_h_out.ptr, _h_in.ptr, &inv_param.param)

# void set_dim(int *)
# void pack_ghost(void **cpuLink, void **cpuGhost, int nFace, QudaPrecision precision)

def computeKSLinkQuda(fatlink, longlink, ulink, inlink, ndarray[double, ndim=1] path_coeff, QudaGaugeParam param):
    _fatlink = _NDArray(fatlink, 2)
    _longlink = _NDArray(longlink, 2)
    _ulink = _NDArray(ulink, 2)
    _inlink = _NDArray(inlink, 2)
    _path_coeff = _NDArray(path_coeff)
    quda.computeKSLinkQuda(_fatlink.ptr, _longlink.ptr, _ulink.ptr, _inlink.ptr, <double *>_path_coeff.ptr, &param.param)

def computeTwoLinkQuda(twolink, inlink, QudaGaugeParam param):
    _twolink = _NDArray(twolink, 2)
    _inlink = _NDArray(inlink, 2)
    quda.computeTwoLinkQuda(_twolink.ptr, _inlink.ptr, &param.param)

def momResidentQuda(mom, QudaGaugeParam param):
    _mom = _NDArray(mom, 2)
    quda.momResidentQuda(_mom.ptr, &param.param)

def computeGaugeForceQuda(mom, sitelink, ndarray[int, ndim=3] input_path_buf, ndarray[int, ndim=1] path_length, ndarray[double, ndim=1] loop_coeff, int num_paths, int max_length, double dt, QudaGaugeParam qudaGaugeParam):
    _mom = _NDArray(mom, 2)
    _sitelink = _NDArray(sitelink, 2)
    _input_path_buf = _NDArray(input_path_buf)
    _path_length = _NDArray(path_length)
    _loop_coeff = _NDArray(loop_coeff)
    return quda.computeGaugeForceQuda(_mom.ptr, _sitelink.ptr, <int ***>_input_path_buf.ptrss, <int *>_path_length.ptr, <double *>_loop_coeff.ptr, num_paths, max_length, dt, &qudaGaugeParam.param)

def computeGaugePathQuda(out, sitelink, ndarray[int, ndim=3] input_path_buf, ndarray[int, ndim=1] path_length, ndarray[double, ndim=1] loop_coeff, int num_paths, int max_length, double dt, QudaGaugeParam qudaGaugeParam):
    _out = _NDArray(out, 2)
    _sitelink = _NDArray(sitelink, 2)
    _input_path_buf = _NDArray(input_path_buf)
    _path_length = _NDArray(path_length)
    _loop_coeff = _NDArray(loop_coeff)
    return quda.computeGaugePathQuda(_out.ptr, _sitelink.ptr, <int ***>_input_path_buf.ptrss, <int *>_path_length.ptr, <double *>_loop_coeff.ptr, num_paths, max_length, dt, &qudaGaugeParam.param)

def computeGaugeLoopTraceQuda(ndarray[double_complex, ndim=1] traces, ndarray[int, ndim=2] input_path_buf, ndarray[int, ndim=1] path_length, ndarray[double, ndim=1] loop_coeff, int num_paths, int max_length, double factor):
    _traces = _NDArray(traces)
    _input_path_buf = _NDArray(input_path_buf)
    _path_length = _NDArray(path_length)
    _loop_coeff = _NDArray(loop_coeff)
    quda.computeGaugeLoopTraceQuda(<double_complex *>_traces.ptr, <int **>_input_path_buf.ptrs, <int *>_path_length.ptr, <double *>_loop_coeff.ptr, num_paths, max_length, factor)

def updateGaugeFieldQuda(gauge, momentum, double dt, int conj_mom, int exact, QudaGaugeParam param):
    _gauge = _NDArray(gauge, 2)
    _momentum = _NDArray(momentum, 2)
    quda.updateGaugeFieldQuda(_gauge.ptr, _momentum.ptr, dt, conj_mom, exact, &param.param)

def staggeredPhaseQuda(gauge_h, QudaGaugeParam param):
    _gauge_h = _NDArray(gauge_h, 2)
    quda.staggeredPhaseQuda(_gauge_h.ptr, &param.param)

def projectSU3Quda(gauge_h, double tol, QudaGaugeParam param):
    _gauge_h = _NDArray(gauge_h, 2)
    quda.projectSU3Quda(_gauge_h.ptr, tol, &param.param)

def momActionQuda(momentum, QudaGaugeParam param):
    _momentum = _NDArray(momentum, 2)
    return quda.momActionQuda(_momentum.ptr, &param.param)

# void* createGaugeFieldQuda(void* gauge, int geometry, QudaGaugeParam* param)
# void saveGaugeFieldQuda(void* outGauge, void* inGauge, QudaGaugeParam* param)
# void destroyGaugeFieldQuda(void* gauge)

def createCloverQuda(QudaInvertParam param):
    quda.createCloverQuda(&param.param)

def computeCloverForceQuda(mom, double dt, x, ndarray[double, ndim=1] coeff, double kappa2, double ck, int nvector, double multiplicity, QudaGaugeParam gauge_param, QudaInvertParam inv_param):
    _mom = _NDArray(mom, 2)
    _x = _NDArray(x, 2)
    _coeff = _NDArray(coeff)
    quda.computeCloverForceQuda(_mom.ptr, dt, _x.ptrs, NULL, <double *>_coeff.ptr, kappa2, ck, nvector, multiplicity, NULL, &gauge_param.param, &inv_param.param)

# void computeStaggeredForceQuda(void *mom, double dt, double delta, void *gauge, void **x, QudaGaugeParam *gauge_param, QudaInvertParam *invert_param)

def computeHISQForceQuda(momentum, double dt, ndarray[double, ndim=1] level2_coeff, ndarray[double, ndim=1] fat7_coeff, w_link, v_link, u_link, quark, int num, int num_naik, ndarray[double, ndim=2] coeff, QudaGaugeParam param):
    _momentum = _NDArray(momentum, 2)
    _level2_coeff = _NDArray(level2_coeff)
    _fat7_coeff = _NDArray(fat7_coeff)
    _w_link = _NDArray(w_link, 2)
    _v_link = _NDArray(v_link, 2)
    _u_link = _NDArray(u_link, 2)
    _quark = _NDArray(quark, 2)
    _coeff = _NDArray(coeff)
    quda.computeHISQForceQuda(_momentum.ptr, dt, <double *>_level2_coeff.ptr, <double *>_fat7_coeff.ptr, _w_link.ptr, _v_link.ptr, _u_link.ptr, _quark.ptrs, num, num_naik, <double **>_coeff.ptrs, &param.param)

def gaussGaugeQuda(unsigned long long seed, double sigma):
    quda.gaussGaugeQuda(seed, sigma)

def gaussMomQuda(unsigned long long seed, double sigma):
    quda.gaussMomQuda(seed, sigma)

def plaqQuda():
    cdef double[3] plaq
    quda.plaqQuda(plaq)
    return plaq

def polyakovLoopQuda(int dir):
    cdef double[2] ploop
    quda.polyakovLoopQuda(ploop, dir)
    return ploop

# void copyExtendedResidentGaugeQuda(void *resident_gauge)

def performWuppertalnStep(h_out, h_in, QudaInvertParam param, unsigned int n_steps, double alpha):
    _h_out = _NDArray(h_out, 1)
    _h_in = _NDArray(h_in, 1)
    quda.performWuppertalnStep(_h_out.ptr, _h_in.ptr, &param.param, n_steps, alpha)

def performGaugeSmearQuda(QudaGaugeSmearParam smear_param, QudaGaugeObservableParam obs_param):
    quda.performGaugeSmearQuda(&smear_param.param, &obs_param.param)

def performWFlowQuda(QudaGaugeSmearParam smear_param, QudaGaugeObservableParam obs_param):
    quda.performWFlowQuda(&smear_param.param, &obs_param.param)

def gaugeObservablesQuda(QudaGaugeObservableParam param):
    quda.gaugeObservablesQuda(&param.param)

def contractQuda(x, y, result, quda.QudaContractType cType, QudaInvertParam param, ndarray[int, ndim=1] X):
    _x = _NDArray(x, 1)
    _y = _NDArray(y, 1)
    _result = _NDArray(result, 1)
    _X = _NDArray(X)
    quda.contractQuda(_x.ptr, _y.ptr, _result.ptr, cType, &param.param, <int *>_X.ptr)

def computeGaugeFixingOVRQuda(gauge, unsigned int gauge_dir, unsigned int Nsteps, unsigned int verbose_interval, double relax_boost, double tolerance, unsigned int reunit_interval, unsigned int stopWtheta, QudaGaugeParam param):
    _gauge = _NDArray(gauge, 2)
    return quda.computeGaugeFixingOVRQuda(_gauge.ptr, gauge_dir, Nsteps, verbose_interval, relax_boost, tolerance, reunit_interval, stopWtheta, &param.param)

def computeGaugeFixingFFTQuda(gauge, unsigned int gauge_dir, unsigned int Nsteps, unsigned int verbose_interval, double alpha, unsigned int autotune, double tolerance, unsigned int stopWtheta, QudaGaugeParam param):
    _gauge = _NDArray(gauge, 2)
    return quda.computeGaugeFixingFFTQuda(_gauge.ptr, gauge_dir, Nsteps, verbose_interval, alpha, autotune, tolerance, stopWtheta, &param.param)

def blasGEMMQuda(arrayA, arrayB, arrayC, quda.QudaBoolean native, QudaBLASParam param):
    _arrayA = _NDArray(arrayA, 1)
    _arrayB = _NDArray(arrayB, 1)
    _arrayC = _NDArray(arrayC, 1)
    quda.blasGEMMQuda(_arrayA.ptr, _arrayB.ptr, _arrayC.ptr, native, &param.param)

def blasLUInvQuda(Ainv, A, quda.QudaBoolean use_native, QudaBLASParam param):
    _Ainv = _NDArray(Ainv, 1)
    _A = _NDArray(A, 1)
    quda.blasLUInvQuda(_Ainv.ptr, _A.ptr, use_native, &param.param)

def flushChronoQuda(int index):
    quda.flushChronoQuda(index)

def newDeflationQuda(QudaEigParam param):
    df_instance = Pointer("void")
    df_instance.set_ptr(quda.newDeflationQuda(&param.param))
    return df_instance

def destroyDeflationQuda(Pointer df_instance):
    quda.destroyDeflationQuda(df_instance.ptr)

cdef class QudaQuarkSmearParam:
    cdef quda.QudaQuarkSmearParam param

    def __init__(self):
        # self.param = quda.QudaQuarkSmearParam()
        pass

    # def __repr__(self):
    #     value = bytearray()
    #     with redirect_stdout(value):
    #         quda.printQudaQuarkSmearParam(&self.param)
    #     return value.decode(sys.stdout.encoding)

    cdef from_ptr(self, quda.QudaQuarkSmearParam *ptr):
        self.param = dereference(ptr)

##%%!! QudaQuarkSmearParam

def performTwoLinkGaussianSmearNStep(h_in, QudaQuarkSmearParam smear_param):
    _h_in = _NDArray(h_in, 1)
    quda.performTwoLinkGaussianSmearNStep(_h_in.ptr, &smear_param.param)



import os
import sys
from setuptools import Extension, setup
from pyquda_pyx import build_pyquda_pyx

if "egg_info" in sys.argv or "dist_info" in sys.argv or "sdist" in sys.argv:
    setup()
elif "QUDA_PATH" in os.environ:
    quda_path = os.path.realpath(os.environ["QUDA_PATH"])
    build_pyquda_pyx(os.path.dirname(__file__), quda_path)
    if os.path.exists(os.path.join(quda_path, "lib", "libquda.so")) or os.path.exists(
        os.path.join(quda_path, "lib", "quda.dll")
    ):
        _STATIC = False
    elif os.path.exists(os.path.join(quda_path, "lib", "libquda.a")) or os.path.exists(
        os.path.join(quda_path, "lib", "quda.lib")
    ):
        _STATIC = True
    else:
        raise FileNotFoundError(f"Cannot find libquda.so or libquda.a in {os.path.join(quda_path, 'lib')}")

    from Cython.Build import cythonize
    import numpy

    extensions = cythonize(
        [
            Extension(
                name="pyquda_comm.pointer",
                sources=["pyquda_comm/src/pointer.pyx"],
                language="c",
            ),
            Extension(
                name="pyquda.quda",
                sources=["pyquda/src/quda.pyx"],
                include_dirs=[os.path.join(quda_path, "include"), numpy.get_include()],
                define_macros=[("NPY_NO_DEPRECATED_API", "NPY_1_7_API_VERSION")],
                library_dirs=[os.path.join(quda_path, "lib")],
                libraries=["quda"],
                extra_link_args=[f"-Wl,-rpath={os.path.join(quda_path, 'lib')}"] if not _STATIC else None,
                language="c",
            ),
            Extension(
                name="pyquda.malloc_quda",
                sources=["pyquda/src/malloc_quda.cpp"],
                include_dirs=[os.path.join(quda_path, "include")],
                library_dirs=[os.path.join(quda_path, "lib")],
                libraries=["quda"],
                extra_link_args=[f"-Wl,-rpath={os.path.join(quda_path, 'lib')}"] if not _STATIC else None,
                language="c++",
            ),
        ],
        language_level="3",
    )

    setup(ext_modules=extensions)
else:
    raise EnvironmentError("QUDA_PATH environment is needed to link against libquda")



from abc import ABC, abstractmethod
from math import exp
from random import Random
from typing import List, Optional, Union

import numpy

from pyquda_comm import getLogger, getArrayBackend
from pyquda_comm.array import arrayRandomGetState, arrayRandomSetState, arrayRandomSeed
from .field import LatticeInfo, LatticeGauge, LatticeMom, LatticeReal
from .quda import (
    QudaGaugeObservableParam,
    gaugeObservablesQuda,
    gaussMomQuda,
    loadGaugeQuda,
    saveGaugeQuda,
    momResidentQuda,
    momActionQuda,
    updateGaugeFieldQuda,
)
from .enum_quda import QudaBoolean, QudaTboundary, QudaVerbosity
from .dirac import WilsonDirac, StaggeredDirac
from .action.abstract import Action, FermionAction, StaggeredFermionAction

nullptr = numpy.empty((0, 0), "<c16")


class Integrator(ABC):
    def __init__(self, n_steps: int) -> None:
        self.n_steps = n_steps

    @abstractmethod
    def integrate(self, updateGauge, updateMom, t: float):
        raise NotImplementedError("The integrator is not implemented yet")


class INT_3G1F(Integrator):
    alpha_ = 0.1
    beta_ = 0.1

    def integrate(self, updateGauge, updateMom, t: float):
        assert self.n_steps % 2 == 0
        dt = t / self.n_steps
        for _ in range(0, self.n_steps, 2):
            updateGauge(dt * (1 / 6 - self.alpha_ / 3))
            updateMom(dt / 3, fermion=False)
            updateGauge(dt * (1 / 3 + self.alpha_ / 3 - self.beta_))
            updateMom(dt, gauge=False)
            updateGauge(dt * (self.alpha_ / 3 + self.beta_))
            updateMom(dt / 3, fermion=False)
            updateGauge(dt * (1 / 3 - self.alpha_ * 2 / 3))
            updateMom(dt / 3, fermion=False)
            updateGauge(dt * (1 / 3 + self.alpha_ * 2 / 3))
            updateMom(dt / 3, fermion=False)
            updateGauge(dt * (1 / 3 - self.alpha_ * 2 / 3))
            updateMom(dt / 3, fermion=False)
            updateGauge(dt * (self.alpha_ / 3 + self.beta_))
            updateMom(dt, gauge=False)
            updateGauge(dt * (1 / 3 + self.alpha_ / 3 - self.beta_))
            updateMom(dt / 3, fermion=False)
            updateGauge(dt * (1 / 6 - self.alpha_ / 3))


class O2Nf1Ng0V(Integrator):
    R"""https://doi.org/10.1016/S0010-4655(02)00754-3
    BAB: Eq.(23), \xi=0"""

    def integrate(self, updateGauge, updateMom, t: float):
        dt = t / self.n_steps
        updateMom(dt / 2)
        for _ in range(self.n_steps - 1):
            updateGauge(dt)
            updateMom(dt)
        updateGauge(dt)
        updateMom(dt / 2)


class O2Nf1Ng0P(Integrator):
    R"""https://doi.org/10.1016/S0010-4655(02)00754-3
    ABA: Eq.(24), \xi=0"""

    def integrate(self, updateGauge, updateMom, t: float):
        dt = t / self.n_steps
        updateGauge(dt / 2)
        for _ in range(self.n_steps - 1):
            updateMom(dt)
            updateGauge(dt)
        updateMom(dt)
        updateGauge(dt / 2)


class O2Nf2Ng0V(Integrator):
    R"""https://doi.org/10.1016/S0010-4655(02)00754-3
    BABAB: Eq.(25), Eq.(31)"""

    lambda_ = 0.1931833275037836

    def integrate(self, updateGauge, updateMom, t: float):
        dt = t / self.n_steps
        updateMom(self.lambda_ * dt)
        for _ in range(self.n_steps - 1):
            updateGauge(dt / 2)
            updateMom((1 - 2 * self.lambda_) * dt)
            updateGauge(dt / 2)
            updateMom(2 * self.lambda_ * dt)
        updateGauge(dt / 2)
        updateMom((1 - 2 * self.lambda_) * dt)
        updateGauge(dt / 2)
        updateMom(self.lambda_ * dt)


class O2Nf2Ng0P(Integrator):
    R"""https://doi.org/10.1016/S0010-4655(02)00754-3
    ABABA: Eq.(32), Eq.(31)"""

    lambda_ = 0.1931833275037836

    def integrate(self, updateGauge, updateMom, t: float):
        dt = t / self.n_steps
        updateGauge(self.lambda_ * dt)
        for _ in range(self.n_steps - 1):
            updateMom(dt / 2)
            updateGauge((1 - 2 * self.lambda_) * dt)
            updateMom(dt / 2)
            updateGauge(2 * self.lambda_ * dt)
        updateMom(dt / 2)
        updateGauge((1 - 2 * self.lambda_) * dt)
        updateMom(dt / 2)
        updateGauge(self.lambda_ * dt)


class O4Nf5Ng0V(Integrator):
    R"""https://doi.org/10.1016/S0010-4655(02)00754-3
    BABABABABAB: Eq.(63), Eq.(71)"""

    rho_ = 0.2539785108410595
    theta_ = -0.03230286765269967
    vartheta_ = 0.08398315262876693
    lambda_ = 0.6822365335719091

    def integrate(self, updateGauge, updateMom, t: float):
        dt = t / self.n_steps
        updateMom(self.vartheta_ * dt)
        for _ in range(self.n_steps - 1):
            updateGauge(self.rho_ * dt)
            updateMom(self.lambda_ * dt)
            updateGauge(self.theta_ * dt)
            updateMom((1 - 2 * (self.lambda_ + self.vartheta_)) * dt / 2)
            updateGauge((1 - 2 * (self.theta_ + self.rho_)) * dt)
            updateMom((1 - 2 * (self.lambda_ + self.vartheta_)) * dt / 2)
            updateGauge(self.theta_ * dt)
            updateMom(self.lambda_ * dt)
            updateGauge(self.rho_ * dt)
            updateMom(2 * self.vartheta_ * dt)
        updateGauge(self.rho_ * dt)
        updateMom(self.lambda_ * dt)
        updateGauge(self.theta_ * dt)
        updateMom((1 - 2 * (self.lambda_ + self.vartheta_)) * dt / 2)
        updateGauge((1 - 2 * (self.theta_ + self.rho_)) * dt)
        updateMom((1 - 2 * (self.lambda_ + self.vartheta_)) * dt / 2)
        updateGauge(self.theta_ * dt)
        updateMom(self.lambda_ * dt)
        updateGauge(self.rho_ * dt)
        updateMom(self.vartheta_ * dt)


class O4Nf5Ng0P(Integrator):
    R"""https://doi.org/10.1016/S0010-4655(02)00754-3
    ABABABABABA: Eq.(72), Eq.(80)"""

    rho_ = 0.2750081212332419
    theta_ = -0.1347950099106792
    vartheta_ = -0.08442961950707149
    lambda_ = 0.3549000571574260

    def integrate(self, updateGauge, updateMom, t: float):
        dt = t / self.n_steps
        updateGauge(self.rho_ * dt)
        for _ in range(self.n_steps - 1):
            updateMom(self.vartheta_ * dt)
            updateGauge(self.theta_ * dt)
            updateMom(self.lambda_ * dt)
            updateGauge((1 - 2 * (self.theta_ + self.rho_)) * dt / 2)
            updateMom((1 - 2 * (self.lambda_ + self.vartheta_)) * dt)
            updateGauge((1 - 2 * (self.theta_ + self.rho_)) * dt / 2)
            updateMom(self.lambda_ * dt)
            updateGauge(self.theta_ * dt)
            updateMom(self.vartheta_ * dt)
            updateGauge(2 * self.rho_ * dt)
        updateMom(self.vartheta_ * dt)
        updateGauge(self.theta_ * dt)
        updateMom(self.lambda_ * dt)
        updateGauge((1 - 2 * (self.theta_ + self.rho_)) * dt / 2)
        updateMom((1 - 2 * (self.lambda_ + self.vartheta_)) * dt)
        updateGauge((1 - 2 * (self.theta_ + self.rho_)) * dt / 2)
        updateMom(self.lambda_ * dt)
        updateGauge(self.theta_ * dt)
        updateMom(self.vartheta_ * dt)
        updateGauge(self.rho_ * dt)


class HMC:
    def __init__(
        self,
        latt_info: LatticeInfo,
        monomials: List[Union[Action, FermionAction, StaggeredFermionAction]],
        integrator: Integrator,
        hmc_inner: Optional["HMC"] = None,
    ) -> None:
        self.latt_info = latt_info
        self.gauge_monomials: List[Action] = [
            monomial for monomial in monomials if not isinstance(monomial, FermionAction)
        ]
        self.fermion_monomials: List[FermionAction] = [
            monomial for monomial in monomials if isinstance(monomial, FermionAction)
        ]
        self.integrator = integrator
        self.hmc_inner = hmc_inner
        if len(self.fermion_monomials) == 0:
            self.is_staggered = False
        else:
            self.is_staggered = isinstance(self.fermion_monomials[0], StaggeredFermionAction)
            for monomial in self.fermion_monomials[1:]:
                if self.is_staggered != isinstance(monomial, StaggeredFermionAction):
                    getLogger().critical(
                        "FermionAction and StaggeredFermionAction cannot be used at the same time", ValueError
                    )
        if not self.is_staggered:
            self._dirac = WilsonDirac(latt_info, 0, 0, 0)
        else:
            self._dirac = StaggeredDirac(latt_info, 0, 0, 0)
        self.gauge_param = self._dirac.gauge_param
        self.obs_param = QudaGaugeObservableParam()
        self.obs_param.remove_staggered_phase = QudaBoolean(self.is_staggered)
        self.fuseFermionAction()
        self.gauge = None  # LatticeGauge(latt_info)
        self.smeared = None  # LatticeGauge(latt_info)
        self.mom = None  # LatticeMom(latt_info)
        self.force = None  # LatticeMom(latt_info)

    def fuseFermionAction(self):
        if self.is_staggered:
            from .action.hisq import HISQAction, MultiHISQAction

            hisq_monomials = [monomial for monomial in self.fermion_monomials if isinstance(monomial, HISQAction)]
            if hisq_monomials != []:
                hisq_monomials = [MultiHISQAction(self.latt_info, hisq_monomials)]
            self.fermion_monomials = hisq_monomials + [
                monomial for monomial in self.fermion_monomials if not isinstance(monomial, HISQAction)
            ]

    def initializeRNG(self, seed: int):
        self.random = Random(seed * self.latt_info.volume + self.latt_info.mpi_rank)
        self.random.setstate(
            (
                self.random.VERSION,
                tuple(self.random.randrange(2**32) if i < 624 else self.random.randrange(625) for i in range(625)),
                None,
            )
        )

    def initialize(self, seed: int, gauge: LatticeGauge, mom: Optional[LatticeMom] = None):
        self.initializeRNG(seed)
        self.loadGauge(gauge)
        if mom is None:
            self.loadMom(LatticeMom(gauge.latt_info))
        else:
            self.loadMom(mom)

    def setFermionVerbosity(self, verbosity: QudaVerbosity):
        for fermion_monomial in self.fermion_monomials:
            fermion_monomial.setVerbosity(verbosity)

    def loadGaugeMomSmeared(self):
        if self.gauge is not None and self.smeared is not None and self.force is not None and self.mom is not None:
            saveGaugeQuda(self.gauge.data_ptrs, self.gauge_param)
            if self.gauge_param.t_boundary == QudaTboundary.QUDA_ANTI_PERIODIC_T:
                self.gauge.setAntiPeriodicT()
            if self.is_staggered:
                self.gauge.staggeredPhase(True)
            self.gauge_param.make_resident_mom = 0
            self.gauge_param.return_result_mom = 1
            momResidentQuda(self.mom.data_ptrs, self.gauge_param)
            self.gauge_param.make_resident_mom = 1
            self.gauge_param.return_result_mom = 0

            self.smeared.data = self.gauge.data
            # self._stoutSmear()
            if self.is_staggered:
                self.smeared.staggeredPhase(False)
            if self.gauge_param.t_boundary == QudaTboundary.QUDA_ANTI_PERIODIC_T:
                self.smeared.setAntiPeriodicT()
            self.gauge_param.use_resident_gauge = 0
            loadGaugeQuda(self.smeared.data_ptrs, self.gauge_param)
            self.gauge_param.use_resident_gauge = 1
            self.force.data = 0

    def loadGaugeMom(self):
        if self.gauge is not None and self.smeared is not None and self.force is not None and self.mom is not None:
            # self._stoutSmearReverse()
            self.mom += self.force

            self.smeared.data = self.gauge.data
            if self.is_staggered:
                self.smeared.staggeredPhase(False)
            if self.gauge_param.t_boundary == QudaTboundary.QUDA_ANTI_PERIODIC_T:
                self.smeared.setAntiPeriodicT()
            self.gauge_param.use_resident_gauge = 0
            loadGaugeQuda(self.smeared.data_ptrs, self.gauge_param)
            self.gauge_param.use_resident_gauge = 1
            momResidentQuda(self.mom.data_ptrs, self.gauge_param)

    def seed(self, state):
        seed = self.random.randrange(2**32)
        backend = getArrayBackend()
        if state is None:
            state = arrayRandomGetState(backend)
            arrayRandomSeed(seed, backend)
        else:
            arrayRandomSetState(state, backend)
        return state

    def samplePhi(self):
        R"""
        \phi\sim e^{-S_f}=e^{-\phi^\dagger\mathcal{M}^{-1}\phi}

        \mathcal{M}^{-\frac{1}{2}}\phi=\eta\sim e^{-\eta^\dagger\eta}
        """
        state = self.seed(None)

        self.loadGaugeMomSmeared()
        for monomial in self.fermion_monomials:
            monomial.sample()
        if self.hmc_inner is not None:
            self.hmc_inner.samplePhi()
        self.loadGaugeMom()

        self.seed(state)

    def momAction(self) -> float:
        return momActionQuda(nullptr, self.gauge_param)

    def gaugeAction(self) -> float:
        action = 0
        for monomial in self.gauge_monomials:
            action += monomial.action()
        if self.hmc_inner is not None:
            action += self.hmc_inner.gaugeAction()
        return action

    def fermionAction(self, use_action_param: bool = False) -> float:
        """
        use_action_param: use rational parameters for fermion action istead of molecular dynamics.
        """
        action = 0
        self.loadGaugeMomSmeared()
        for monomial in self.fermion_monomials:
            action += monomial.action()  # if not use_action_param else monomial.actionFA()
        if self.hmc_inner is not None:
            action += self.hmc_inner.fermionAction()
        self.loadGaugeMom()
        return action

    def gaugeForce(self, dt: float):
        for monomial in self.gauge_monomials:
            monomial.force(dt, self.force)

    def fermionForce(self, dt: float):
        self.loadGaugeMomSmeared()
        for monomial in self.fermion_monomials:
            monomial.force(dt, self.force)
        self.loadGaugeMom()

    def updateGauge(self, dt: float):
        updateGaugeFieldQuda(nullptr, nullptr, dt, False, True, self.gauge_param)

    def updateMom(self, dt: float, gauge: bool = True, fermion: bool = True):
        if gauge:
            self.gaugeForce(dt)
        if fermion:
            self.fermionForce(dt)

    def loadGauge(self, gauge: LatticeGauge):
        gauge_in = gauge.copy()
        if self.is_staggered:
            gauge_in.staggeredPhase(False)
        if self.gauge_param.t_boundary == QudaTboundary.QUDA_ANTI_PERIODIC_T:
            gauge_in.setAntiPeriodicT()
        self.gauge_param.use_resident_gauge = 0
        loadGaugeQuda(gauge_in.data_ptrs, self.gauge_param)
        self.gauge_param.use_resident_gauge = 1

    def saveGauge(self, gauge: LatticeGauge):
        saveGaugeQuda(gauge.data_ptrs, self.gauge_param)
        if self.gauge_param.t_boundary == QudaTboundary.QUDA_ANTI_PERIODIC_T:
            gauge.setAntiPeriodicT()
        if self.is_staggered:
            gauge.staggeredPhase(True)

    def loadMom(self, mom: LatticeMom):
        momResidentQuda(mom.data_ptrs, self.gauge_param)

    def saveMom(self, mom: LatticeMom):
        self.gauge_param.make_resident_mom = 0
        self.gauge_param.return_result_mom = 1
        momResidentQuda(mom.data_ptrs, self.gauge_param)
        self.gauge_param.make_resident_mom = 1
        self.gauge_param.return_result_mom = 0
        momResidentQuda(mom.data_ptrs, self.gauge_param)  # keep momResident

    def gaussMom(self):
        gaussMomQuda(self.random.randrange(2**32), 1.0)

    def projectSU3(self, tol: float):
        gauge = LatticeGauge(self.latt_info)
        self.saveGauge(gauge)
        gauge.projectSU3(tol)
        self.loadGauge(gauge)

    def integrate(self, t: float, project_tol: float = 0.0):
        self.integrator.integrate(
            self.updateGauge if self.hmc_inner is None else self.hmc_inner.integrate, self.updateMom, t
        )
        if project_tol > 0.0:
            self.projectSU3(project_tol)

    def accept(self, delta_s: float):
        return self.latt_info.mpi_comm.bcast(self.random.random() < exp(min(-delta_s, 0)))

    def plaquette(self):
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_plaquette = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.plaquette

    def polyakovLoop(self):
        self.obs_param.compute_polyakov_loop = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_polyakov_loop = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.ploop

    def energy(self):
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.energy

    def qcharge(self):
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_qcharge = QudaBoolean.QUDA_BOOLEAN_FALSE
        return self.obs_param.qcharge

    def qchargeDensity(self):
        qcharge_density = LatticeReal(self.latt_info)
        self.obs_param.qcharge_density = qcharge_density.data_void_ptr
        self.obs_param.compute_qcharge_density = QudaBoolean.QUDA_BOOLEAN_TRUE
        gaugeObservablesQuda(self.obs_param)
        self.obs_param.compute_qcharge_density = QudaBoolean.QUDA_BOOLEAN_TRUE
        return qcharge_density
